import json
from typing import Callable, Any, Tuple, Union, Dict

from ..file_manager import FileManager

PYTHON_SERIALIZABLE_TYPES = Union[dict, list, tuple, str, int, float, bool, None]


class TqJSONEncoder(json.JSONEncoder):
    """
    The default JSONEncoder of tq_json package.
    You can set serialize function through set_class_serialize_function, so that TqJSONEncoder can convert the correct type of class to the Python serializable object.
    Also, you don't have to self-make Encoder again.
    See set_class_serialize_function function for more details.
    """
    # 存储设置的类对应的字典化方法，只需要得到可序列化对象即可，不用得到 json 字符串。
    class_serialize_functions: Dict[Union[type, Tuple[type, ...]], Callable[[Any], PYTHON_SERIALIZABLE_TYPES]] = dict()

    @classmethod
    def set_class_serialize_function(cls, class_type: Union[type, Tuple[type, ...]],
                                     dictify_function: Callable[[Any], PYTHON_SERIALIZABLE_TYPES]):
        """
        Set class_serialize_function.
        In condition that the object is instance of class_type, the dump methods in the tq_json package will call class_serialize_function to create Python serializable object through given object.
        :param class_type: the type of the class
        :param dictify_function: the function that will create dict through object. Notice: The keys of dict should be the property names of the object.
        """
        cls.class_serialize_functions[class_type] = dictify_function

    def default(self, obj):
        for key, value in self.class_serialize_functions.items():
            if isinstance(obj, key):
                return value(obj)  # 最终方法返回的结果是，可以 json 序列化的对象
        return json.JSONEncoder.default(self, obj)


class TqJSONDecoder(json.JSONDecoder):
    """
    The default JSONDecoder of tq_json package.
    You can set objectify function through set_class_objectify_function, so that TqJSONDecoder can convert correct dict to the specified class.
    Also, you don't have to self-make Decoder again.
    See set_class_objectify_function function for more details.
    """
    # 存储设置的类对应的对象化方法。key 是类的属性名元组，value 是对象化方法。因为 json 反序列化为字典后，属性值是唯一标识了，那么就是用属性名元组进行区分。
    class_objectify_functions: Dict[Tuple[str, ...], Callable[[dict], Any]] = dict()

    def __init__(self, *args, **kwargs):
        # 使用 object_hook 实现自定义的解码逻辑
        super().__init__(object_hook=self.object_hook, *args, **kwargs)

    @classmethod
    def set_class_objectify_function(cls, class_property_names: Tuple[str, ...],
                                     objectify_function: Callable[[dict], Any]):
        """
        Set class_objectify_function.
        In condition that the keys of dict are exactly same as the property names of class, the load methods in the tq_json package will call class_objectify_function to create class object through dict.
        :param class_property_names: tuple of class's property names.
        :param objectify_function: the function that will create class object through dict
        """
        cls.class_objectify_functions[class_property_names] = objectify_function

    def object_hook(self, obj: dict):
        for key, value in self.class_objectify_functions.items():
            class_property_names = sorted(list(key))
            obj_property_names = sorted(list(obj.keys()))
            if class_property_names == obj_property_names:
                return value(obj)
        return obj


def dump_json(path, json_data, json_encoder=TqJSONEncoder):
    """
    with-open 配合 json.dump 的普通 json 转换以及文本输出.
    :param path: 输出 json 文本的文件路径
    :param json_data: 转换成 json 文本的对象
    :param json_encoder: 自定义 JSONEncoder
    """
    with open(path, 'w', encoding='utf-8') as f:
        json.dump(json_data, f, cls=json_encoder, indent=4, ensure_ascii=False)


def dump_json_safe(path, json_data, json_encoder=TqJSONEncoder):
    """
    将对象转换成 json 文本并安全地写入 json 文件中. 安全，即若 json 文件的父目录不存在，则会进行创建。
    :param path: 输出 json 文本的文件路径
    :param json_data: 转换成 json 文本的对象
    :param json_encoder: 自定义 JSONEncoder
    """
    with FileManager(path, 'w', encoding='utf-8') as f:
        json.dump(json_data, f, cls=json_encoder, indent=4, ensure_ascii=False)


def dumps_json(json_data, json_encoder=TqJSONEncoder):
    """
    将对象转换得到 json 文本
    :param json_data: 转换成 json 文本的对象
    :param json_encoder: 自定义 JSONEncoder
    :return: 转换的 json 文本
    """
    return json.dumps(json_data, cls=json_encoder, indent=4, ensure_ascii=False)


def load_json(path: str, encoding='utf-8', json_decoder=TqJSONDecoder):
    """
    with-open 配合 json.load 的普通 json 文本读取并转换得到对象.
    :param path: json 文件的路径
    :param encoding: json 文件中文本的编码
    :param json_decoder: 自定义 JSONDecoder
    :return: json 文本读取并转换得到的对象
    """
    with open(path, 'r', encoding=encoding) as f:
        return json.load(f, cls=json_decoder)


def load_json_safe(path: str, encoding='utf-8', json_decoder=TqJSONDecoder):
    """
    从 json 文件中安全地读取 json 文本并得到转换的对象. 安全，即若 json 文件不存在，则会创建一个空白的 json 文件并读取.
    :param path: json 文件的路径
    :param encoding: json 文件中文本的编码
    :param json_decoder: 自定义 JSONDecoder
    :return: json 文本读取并转换得到的对象
    """
    with FileManager(path, 'r', encoding=encoding, initial_data='') as f:
        return json.load(f, cls=json_decoder)


def loads_json(json_string, json_decoder=TqJSONDecoder):
    """
    将 json 文本转换得到对象
    :param json_string: 需要转换的 json 文本
    :param json_decoder: 自定义 JSONDecoder
    :return: json 文本转换得到的对象
    """
    return json.loads(json_string, cls=json_decoder)



from typing import Optional

from .timer import Timer

# func_timeout 用起来有问题，实际用 timer 计时用了 0.03-0.04s，但是设定小于该值的 timeout 不会抛出异常，但进行调试到是正常。
import func_timeout as ft  # https://github.com/kata198/func_timeout

func_set_timeout = ft.func_set_timeout  # Decorator
FunctionTimedOut = ft.FunctionTimedOut
StoppableThread = ft.StoppableThread


def func_timeout(timeout: float, func, args: tuple = (), kwargs: Optional[dict] = None, timer: Timer = None):
    """ Copy from func_timeout.func_timeout
    Runs the given function for up to #timeout# seconds. Raises any exceptions #func# would raise, returns what #func# would return (unless timeout is exceeded), in which case it raises FunctionTimedOut
    :param timeout: Maximum number of seconds to run #func# before terminating. timeout must be positive, otherwise it will raise ValueError.
    :param func: The function to call.
    :param args: Any ordered arguments to pass to the function. Default: ().
    :param kwargs: Keyword arguments to pass to the function. Default: None.
    :param timer: if specified, timer will be used to keep track of how much time the #func# uses.
    :raises FunctionTimedOut:
        Raise FunctionTimedOut, if #timeout# is exceeded, otherwise anything #func# could raise will be raised
        Note: If the timeout is exceeded, FunctionTimedOut will be raised within the context of the called function every two seconds until it terminates,
        but will not block the calling thread (a new thread will be created to perform the join). If possible, you should try/except FunctionTimedOut
        to return cleanly, but in most cases it will 'just work'.
    :returns: The return value that #func# gives
    """
    if timeout <= 0:
        raise ValueError('timeout must be positive.')
    if timer:
        with timer:
            result = ft.func_timeout(timeout, func, args, kwargs)
    else:
        result = ft.func_timeout(timeout, func, args, kwargs)
    return result



import pickle

from ..file_manager import FileManager


def load(path: str, initial_data=None):
    """
    Load pickle file to Python object.
    :param path: pickle file path. file suffix is '.pkl'.
    :param initial_data: if file doesn't exist, it'll initialize it with initial data. Default is empty list.
    """
    if initial_data is None:
        initial_data = []
    with FileManager(path, mode='rb', encoding=None, initial_data=pickle.dumps(initial_data)) as fr:
        return pickle.load(fr)


def dump(path: str, data):
    """
    Dump Python object to pickle file.
    :param path: pickle file path. file suffix is '.pkl'.
    :param data: Python object to dump.
    """
    with FileManager(path, mode='wb', encoding=None) as fw:
        pickle.dump(data, fw)



# import 子包
from . import file_manager, tq_json, tq_pickle, time_util, tq_logger, tq_sqlite3
from .singleton import SingletonTypeThreadSafe

# 提供统一对外API，通过 from utils import * 方式使用
__all__ = ['file_manager', 'tq_json', 'tq_pickle', 'SingletonTypeThreadSafe', 'time_util', 'tq_logger', 'tq_sqlite3']



from .timer import Timer, TimerError
from .func_time_util import func_timeout, func_set_timeout, FunctionTimedOut, StoppableThread

__all__ = ['Timer', 'TimerError', 'func_timeout', 'func_set_timeout', 'FunctionTimedOut', 'StoppableThread']



import enum
import functools
import os
import time
from typing import Any, Callable

from ..file_manager import FileManager


class LoggingLevel(enum.Enum):
    DEBUG = 10, 'DEBUG'
    INFO = 20, 'INFO'
    WARNING = 30, 'WARNING'
    ERROR = 40, 'ERROR'
    CRITICAL = 50, 'CRITICAL'


class TQRunLogger:
    """
    用于记录每单次运行的日志输出.
    支持 实例, Decorator 和 Context Manager 的使用方式。
    """
    __LOGGING_MESSAGE_FORMAT = '[{time}][{level}]: {message}\n'
    __LOGGING_DIR_FORMAT = '%Y-%m-%d-%H-%M-%S'
    __LOGGING_TIME_FORMAT = '%Y%m%d-%H:%M:%S'

    def __init__(self, logging_output_dir: str, logging_level: LoggingLevel = LoggingLevel.DEBUG):
        self.__run_logging_output_dir = os.path.join(logging_output_dir,
                                                     time.strftime(self.__LOGGING_DIR_FORMAT, time.localtime()))
        logger_file_path = os.path.join(self.__run_logging_output_dir, 'logging.log')
        file_manager = FileManager(logger_file_path, 'at')
        self.__logger_file = file_manager.open()
        self.__logging_level_value = logging_level.value[0]

    def __logging_message(self, level: LoggingLevel, message: str):
        level_value, level_str = level.value
        if level_value < self.__logging_level_value:
            return
        self.__logger_file.write(
            self.__get_format_message(level_str=level_str, message=message))

    def __get_format_message(self, level_str: str, message: str):
        time_str = time.strftime(self.__LOGGING_TIME_FORMAT, time.localtime())
        return self.__LOGGING_MESSAGE_FORMAT.format(level=level_str, time=time_str, message=message)

    def debug(self, message: str):
        self.__logging_message(level=LoggingLevel.DEBUG, message=message)

    def info(self, message: str):
        self.__logging_message(level=LoggingLevel.INFO, message=message)

    def warning(self, message: str):
        self.__logging_message(level=LoggingLevel.WARNING, message=message)

    def error(self, message: str):
        self.__logging_message(level=LoggingLevel.ERROR, message=message)

    def critical(self, message: str):
        self.__logging_message(level=LoggingLevel.CRITICAL, message=message)

    def logging_output_file(self, file_name: str, file_data: Any, file_save_function: Callable[[str, Any], None]):
        """
        日志输出文件
        :param file_name: 文件全称
        :param file_data: 文件数据
        :param file_save_function: 文件保存处理函数. params(file_path:str, file_data: Any)
        """
        filepath = os.path.join(self.__run_logging_output_dir, file_name)
        abs_filepath = os.path.abspath(filepath)
        file_dir = os.path.dirname(abs_filepath)
        if not os.path.exists(file_dir):
            os.makedirs(file_dir)
        file_save_function(abs_filepath, file_data)
        self.info('save file at {}'.format(abs_filepath))

    def __enter__(self):
        """Support using logger as a context manager"""
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.__logger_file.close()

    def __call__(self, func):
        """Support using logger as a decorator"""

        @functools.wraps(func)
        def wrapper_logger(*args, **kwargs):
            with self as logger:
                return func(logger, *args, **kwargs)

        return wrapper_logger



"""
使用方法：
    class Singleton(metaclass=SingletonTypeThreadSafe):
        def __init__(self, value):
            pass
如此，即可是的定义的 Singleton 类为单例模式. SingletonTypeThreadSafe是线程安全的。
"""

from .singleton_meta_class import SingletonTypeThreadSafe

__all__ = ['SingletonTypeThreadSafe']



import threading


class SingletonTypeThreadSafe(type):
    # _instance = None  # 只存储一个类的实例，即仅支持一个单例类
    _instances = {}  # 支持多个单例类的实例管理
    _instance_lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        with SingletonTypeThreadSafe._instance_lock:
            if cls not in cls._instances:
                cls._instances[cls] = super(SingletonTypeThreadSafe, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



"""
def test_open_file():  # context manager usage
    print()
    # loading a file
    # 路径以测试环境为主，E:\Workspace\Pycharm\TQUtils\test
    with FileManager(add_base_dir(r'test.txt'), 'w') as f:
        f.write('Test')
    print(f.closed)  # Output:true


def test_use_file_manager_through_instance():  # instance usage
    fm = FileManager(add_base_dir(r'test1.txt'), 'r')
    f = fm.open()
    print(f.read())
    fm.close()
    # or
    # f.close()
"""
from .FileManager import FileManager, InvalidFileNameException, UnsupportedPlatformException, ModeErrorException, \
    PermissionErrorException, FileNotFoundException, is_valid_filename

__all__ = ['FileManager', 'UnsupportedPlatformException', 'ModeErrorException', 'PermissionErrorException',
           'InvalidFileNameException', 'FileNotFoundException', 'is_valid_filename']



import os
import platform
import re
from typing import Callable, Optional, Union


class ModeErrorException(Exception):
    def __init__(self, error_info):
        super().__init__(self)
        self.error_info = error_info

    def __str__(self):
        return self.error_info


class PermissionErrorException(Exception):
    def __init__(self, error_info):
        super().__init__(self)
        self.error_info = error_info

    def __str__(self):
        return self.error_info


class FileNotFoundException(Exception):
    def __init__(self, error_info):
        super().__init__(self)
        self.error_info = error_info

    def __str__(self):
        return self.error_info


class UnsupportedPlatformException(Exception):
    def __init__(self, error_info):
        super().__init__(self)
        self.error_info = error_info

    def __str__(self):
        return self.error_info


class InvalidFileNameException(Exception):
    def __init__(self, error_info):
        super().__init__(self)
        self.error_info = error_info

    def __str__(self):
        return self.error_info


def is_valid_filename(filename: str) -> bool:
    """
    验证文件名的有效性，区分 Windows 和 Linux。

    :param filename: 文件名
    :return: 如果文件名有效返回 True，否则返回 False
    """
    # 检查文件名长度
    if not (1 <= len(filename) <= 255):
        return False

    # 获取操作系统
    os_name = platform.system()

    # 定义非法字符和保留文件名
    if os_name == 'Windows':
        # Windows 系统的非法字符
        illegal_chars = r'[<>:"/\\|?*]'
        reserved_names = {'CON', 'PRN', 'AUX', 'NUL', 'COM1', 'COM2', 'COM3', 'COM4', 'COM5', 'COM6', 'COM7', 'COM8',
                          'COM9', 'LPT1', 'LPT2', 'LPT3', 'LPT4', 'LPT5', 'LPT6', 'LPT7', 'LPT8', 'LPT9'}

        # 检查非法字符
        if re.search(illegal_chars, filename):
            return False

        # 检查是否是保留文件名
        name_only = os.path.splitext(filename)[0].upper()
        if name_only in reserved_names:
            return False

    elif os_name == 'Linux':
        # Linux 系统的非法字符（几乎没有限制）
        # 这里可以根据需要添加特定的检查
        if '/' in filename:
            return False

    else:
        raise UnsupportedPlatformException("Unsupported operating system.")

    return True


class FileManager:
    """
    主要通过 Context Manager 的方式使用，也可以通过实例使用: f.open() 获取 TextIOWrapper。
    """

    def __init__(self, path: Union[int, str], mode: str, buffering: int = -1, encoding: Optional[str] = 'utf-8',
                 errors: str = None, newline: str = None, closefd: bool = True,
                 opener: Optional[Callable[[str, int], int]] = None, initial_data: Union[str, bytes] = None):
        """
        :param path: file path to open
        :param mode: file open mode.
            e.g. 'r', 'rb', 'rt', 'r+', 'rb+', 'rt+',
                 'w', 'wb', 'wt', 'w+', 'wb+', 'wt+',
                 'a', 'ab', 'at',
                 'x', 'xb', 'xt', 'x+', 'xb+', 'xt+'
        :param buffering: buffer size
        :param encoding: file encoding mode. binary mode doesn't take an encoding argument.
        :param errors: default None. Set 'ignore' to ignore the decode error and output illegal characters.
        :param newline: Line feed processing mode, None(default), ''(blank string), '\n', '\r\n', '\r'.
        :param closefd: use file descriptor to close file. turn on it when use a file descriptor to open file.
        :param opener: open file opener.
        :param initial_data: It'll automatically create the file and initialize it with initial data, if it doesn't exist.
        """
        self.path = path
        self.mode = mode
        self.buffering = buffering
        self.encoding = encoding
        self.errors = errors
        self.newline = newline
        self.closefd = closefd
        self.opener = opener
        self.initial_data = initial_data
        self.file = None
        self.modes_r = ['r', 'rb', 'rt', 'r+', 'rb+', 'rt+']
        self.modes_w = ['w', 'wb', 'wt', 'w+', 'wb+', 'wt+']
        self.modes_a = ['a', 'ab', 'at']
        self.modes_x = ['x', 'xb', 'xt', 'x+', 'xb+', 'xt+']

    def open(self):
        abspath = os.path.abspath(self.path)
        dirs = os.path.dirname(abspath)  # 获得路径的目录
        os.makedirs(dirs, exist_ok=True)  # 创建所有父文件夹
        file_name = os.path.basename(abspath)  # 获取文件名称
        if not is_valid_filename(file_name):
            raise InvalidFileNameException('Invalid filename.')

        if self.mode in self.modes_r:
            # 读文件需要判断文件是否存在，不存在则创建
            if not os.path.exists(abspath):
                if isinstance(self.initial_data, str):
                    mode = self.modes_w[0]
                    encoding = 'utf-8'
                else:
                    mode = self.modes_w[1]
                    encoding = None
                f = open(abspath, mode, encoding=encoding)
                f.write(self.initial_data)
                f.close()
        elif self.mode in self.modes_x:  # 想想还是不对创建写进行处理（删掉已存在的文件或改为覆盖写）,就让它自己抛出异常吧
            pass
        elif self.mode not in self.modes_w and self.mode not in self.modes_a:
            raise ModeErrorException('the input mode is wrong')

        try:
            self.file = open(self.path, self.mode, self.buffering, self.encoding, self.errors, self.newline,
                             self.closefd, self.opener)
        except PermissionError:
            raise PermissionErrorException('Insufficient permissions, please run as administrator.')
        return self.file

    def close(self):
        if not self.file.close:
            self.file.close()

    def __enter__(self):
        return self.open()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()



"""
Timer 用于计时的类、上下文管理器 和 修饰器。详细用法见对应的测试代码。
"""
import functools
import time
from dataclasses import dataclass, field
from typing import ClassVar, Any, Dict


class TimerError(Exception):
    """A custom exception used to report errors in use of Timer class"""


@dataclass
class Timer:
    """
    Timer class, context manager and decorator.
    :property: name: the name of the timer. If you use timer through context manager, the timer must be named.
    :property: text: the string format of time report.
    :property: logger: the logging function.
    """
    __timers: ClassVar[Dict[str, float]] = {}
    name: str = None
    text: str = "{} elapsed time: {:0.4f} seconds"
    logger: Any = print
    __start_time: Any = field(default=None, init=False, repr=False)
    __last_start_time: Any = field(default=None, init=False, repr=False)

    def __post_init__(self):
        """Initialization: add timer to dict of timers. will be called after __init__ """
        if self.name:  # 为每个 timer 实例初始化对应名称的消耗时间和开始时间（没有指定名称不支持保留消耗时间）.
            self.__timers.setdefault(self.name, 0)

    def start(self):
        """Start a new timer"""
        if self.__start_time is not None:
            raise TimerError(f"Timer is running. Use .stop() to stop it")

        self.__start_time = time.perf_counter()
        self.__last_start_time = time.localtime()  # 记录最新地开始计时时间

    def stop(self) -> float:
        """
        Stop the timer, report and return the elapsed time(seconds).
        :return: elapsed time in seconds.
        """
        if self.__start_time is None:
            raise TimerError(f"Timer is not running. Use .start() to start it")

        elapsed_time = time.perf_counter() - self.__start_time
        self.__start_time = None

        if self.logger:
            self.logger(self.text.format(self.name if self.name else 'Timer', elapsed_time))
        if self.name:  # 如果有命名，则会累计时间
            self.__timers[self.name] += elapsed_time  # 累计这次的时间，并还是存储在 timers 中。

        return elapsed_time

    @property
    def last_start_time(self) -> str:
        """ return the last start time in '%Y-%m-%d %H:%M:%S' format. """
        return time.strftime('%Y-%m-%d %H:%M:%S', self.__last_start_time)

    @classmethod
    def getTimerNames(cls):
        """
        Get Timer names.
        :return list: the str names of the timers
        """
        return cls.__timers.keys()

    @classmethod
    def popTime(cls, name, default=None) -> float:
        """
        Get Timer's time and delete timer.
        :param name: Timer name.
        :param default: default value, if timer not exists.
        :return: time elapsed (seconds), or default value if timer not exists.
        """
        return cls.__timers.pop(name, default)

    def __enter__(self):
        """Start a new timer as a context manager"""
        self.start()
        return self

    def __exit__(self, *exc_info):
        """Stop the context manager timer"""
        self.stop()

    def __call__(self, func):
        """Support using Timer as a decorator"""

        @functools.wraps(func)
        def wrapper_timer(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return wrapper_timer



import sqlite3
from abc import abstractmethod
from typing import Optional


class Sqlite3Template:
    def __init__(self, db_file):
        self.db_file = db_file
        self._last_result: Optional[dict] = None
        self.__conn: Optional[sqlite3.Connection] = None

    def __connect(self):
        """
        创建数据库链接，db_file(数据库名称)在创建对象时传入
        """
        self.__conn = sqlite3.connect(self.db_file)  # 如果db文件不存在，connect函数将创建不存在的db文件
        self.__conn.row_factory = sqlite3.Row

    def do_operate(self, query_statement):
        """
        :param query_statement: SQL 语句
        e.g.
        do_operate('CREATE TABLE IF NOT EXISTS users (id INT PRIMARY KEY, name VARCHAR(100) NOT NULL );')
        do_operate('INSERT INTO users VALUES (2, "Tommy")')
        """
        cursor = None
        try:
            self.__connect()
            cursor = self.__conn.execute(query_statement)
            self._last_result = {'rowcount': cursor.rowcount, 'fetchall': cursor.fetchall()}
            self.__conn.commit()
        finally:
            if cursor is not None:
                cursor.close()
            if self.__conn is not None:
                self.__conn.close()

    def do_operate_with_parameters(self, query_statement, parameters=None):
        """
        :param query_statement: SQL 语句
        :param parameters: 如果有很多值需要传入，在字符串中直接拼接SQL是不安全的，可以使用占位符解决.
        e.g.
        do_operate_with_parameters('INSERT INTO users VALUES (?, ?)', (3, "Jessica"))
        """
        cursor = None
        try:
            self.__connect()
            cursor = self.__conn.execute(query_statement, parameters)
            self._last_result = {'rowcount': cursor.rowcount, 'fetchall': cursor.fetchall()}
            self.__conn.commit()
        finally:
            if cursor is not None:
                cursor.close()
            if self.__conn is not None:
                self.__conn.close()

    def do_operates_with_parameters(self, query_statement, parameters=None):
        """
        批量操作，如批量插入
        data = [(1, "Ridesharing"), (2, "Water Purifying"), (3, "Forensics"), (4, "Botany")]
        do_operates_with_parameters("INSERT INTO projects VALUES(?, ?)", data)
        """
        cursor = None
        try:
            self.__connect()
            cursor = self.__conn.executemany(query_statement, parameters)
            self._last_result = {'rowcount': cursor.rowcount, 'fetchall': cursor.fetchall()}
            self.__conn.commit()
        finally:
            if cursor is not None:
                cursor.close()
            if self.__conn is not None:
                self.__conn.close()

    @abstractmethod
    def format_result(self):  # 返回格式化的结果，由子类实现
        raise NotImplementedError()

    @abstractmethod
    def get_result(self):  # 返回列表(行之间)的结果，由子类实现
        raise NotImplementedError()


class ExecuteUpdate(Sqlite3Template):
    """ ExecuteUpdate用于执行update、delete、insert语句等等，返回一个整数 """

    def format_result(self):
        """ 返回数据样式的举例: 'n'(为执行返回的整数) """
        output = str(self._last_result['rowcount'])
        return output

    def get_result(self):
        """ 返回数据样式的举例: n(为执行返回的整数) """
        output = int(self._last_result['rowcount'])
        return output


class ExecuteQuery(Sqlite3Template):
    """ ExecuteQuery用于执行SELECT语句，用于产生单个结果集 """

    def format_result(self):
        """
        返回数据样式的举例: "1,202031990141,张三\n2,202031990121,李四"
        """
        output = []
        # 在父类的operate_with_parameters函数中，操作的结果存储在result中
        for row_dict in self._last_result['fetchall']:  # fetchall函数，从结果集中获取所有行(list)
            lst = []  # 存放每行的数据
            for k in dict(row_dict):  # 通过for循环遍历一行的结果，将每个列值存放到lst中
                lst.append(str(row_dict[k]))
            output.append(', '.join(lst))
        return '\n'.join(output)  # 每行直接通过两个回车连接，最后返回的是字符串类型

    def get_result(self):
        """
        返回数据样式的举例:
        [{'id':1,'studentId':'202031990141','name':'张三'},{'id':2,'studentId':'202031990121','name':'李四'}]
        """
        output = []
        for row_dict in self._last_result['fetchall']:
            output.append(dict(row_dict))  # 行之间以列表的形式存储，列之间为字典形式存储
        return output



from .tq_json import dump_json, dumps_json, loads_json, load_json, dump_json_safe, load_json_safe, TqJSONDecoder, TqJSONEncoder

__all__ = ['dumps_json', 'dump_json', 'loads_json', 'load_json', 'dump_json_safe', 'load_json_safe', 'TqJSONDecoder', 'TqJSONEncoder']



"""
Usage of TQRunLogger

def case(logger):
    logger.info('asd')
    logger.warning('123')
    logger.error('456')

    def save_function(file_path, file_data):
        with FileManager(file_path, 'w') as f:
            f.write(file_data)

    logger.logging_output_file('asd-{}.txt'.format(hash(random.random())), '123123', save_function)


def test_logger_instance():  # instance usage
    logger = TQRunLogger(LOGGER_TEST_DATA_DIR)
    case(logger)


def test_logger_decorator():  # decorator usage
    @TQRunLogger(LOGGER_TEST_DATA_DIR)
    def decorator_method(logger):
        case(logger)

    decorator_method()

def test_logger_context_manager():  # context manager usage
    with TQRunLogger(LOGGER_TEST_DATA_DIR) as logger:
        case(logger)
"""
from .logger import TQRunLogger, LoggingLevel

__all__ = ['TQRunLogger', 'LoggingLevel']
